import pytz
import datetime
import pandas

top_bound = pandas.to_datetime(datetime.datetime(year=2020, month=12, day=31, tzinfo=pytz.timezone('EST')))
low_bound = pandas.to_datetime(datetime.datetime(year=2019, month=1, day=1, tzinfo=pytz.timezone('EST')))
zero_diff = pandas.Timedelta(days=0)


class BusyDayExchange:

    @staticmethod
    def y2019():
        holiday = [datetime.datetime(year=2019, month=1, day=1, tzinfo=pytz.timezone('EST')),
                   datetime.datetime(year=2019, month=1, day=21, tzinfo=pytz.timezone('EST')),
                   datetime.datetime(year=2019, month=2, day=18, tzinfo=pytz.timezone('EST')),
                   datetime.datetime(year=2019, month=4, day=19, tzinfo=pytz.timezone('EST')),
                   datetime.datetime(year=2019, month=5, day=27, tzinfo=pytz.timezone('EST')),
                   datetime.datetime(year=2019, month=7, day=4, tzinfo=pytz.timezone('EST')),
                   datetime.datetime(year=2019, month=9, day=2, tzinfo=pytz.timezone('EST')),
                   datetime.datetime(year=2019, month=11, day=28, tzinfo=pytz.timezone('EST')),
                   datetime.datetime(year=2019, month=12, day=25, tzinfo=pytz.timezone('EST'))]
        cutday = [datetime.datetime(year=2019, month=7, day=3, tzinfo=pytz.timezone('EST')),
                  datetime.datetime(year=2019, month=11, day=29, tzinfo=pytz.timezone('EST')),
                  datetime.datetime(year=2019, month=12, day=24, tzinfo=pytz.timezone('EST'))]
        holiday = pandas.Series(pandas.to_datetime(holiday))
        cutday = pandas.Series(pandas.to_datetime(cutday))
        return holiday, cutday

    @staticmethod
    def y2020():
        holiday = [datetime.datetime(year=2020, month=1, day=1, tzinfo=pytz.timezone('EST')),
                   datetime.datetime(year=2020, month=1, day=20, tzinfo=pytz.timezone('EST')),
                   datetime.datetime(year=2020, month=2, day=17, tzinfo=pytz.timezone('EST')),
                   datetime.datetime(year=2020, month=4, day=10, tzinfo=pytz.timezone('EST')),
                   datetime.datetime(year=2020, month=5, day=25, tzinfo=pytz.timezone('EST')),
                   datetime.datetime(year=2020, month=7, day=3, tzinfo=pytz.timezone('EST')),
                   datetime.datetime(year=2020, month=9, day=7, tzinfo=pytz.timezone('EST')),
                   datetime.datetime(year=2020, month=11, day=26, tzinfo=pytz.timezone('EST')),
                   datetime.datetime(year=2020, month=12, day=25, tzinfo=pytz.timezone('EST'))]
        cutday = [datetime.datetime(year=2020, month=11, day=27, tzinfo=pytz.timezone('EST')),
                  datetime.datetime(year=2020, month=12, day=24, tzinfo=pytz.timezone('EST'))]
        holiday = pandas.Series(pandas.to_datetime(holiday))
        cutday = pandas.Series(pandas.to_datetime(cutday))
        return holiday, cutday

    @staticmethod
    def check_range(start_date, end_date):
        if start_date < low_bound or end_date > top_bound:
            raise Exception("Not Implemented")
        if start_date > end_date:
            raise ValueError("start_date should be lesser than end_date")
        return None

    @staticmethod
    def get_ranges(start_date, end_date):
        BusyDayExchange.check_range(start_date, end_date)
        if end_date > pandas.to_datetime(datetime.datetime(year=2020, month=1, day=1, tzinfo=pytz.timezone('EST'))):
            y2020_holiday, y2020_cutday = BusyDayExchange.y2020()
        else:
            y2020_holiday, y2020_cutday = pandas.Series([], dtype=object), pandas.Series([], dtype=object)
        if start_date < pandas.to_datetime(datetime.datetime(year=2020, month=1, day=1, tzinfo=pytz.timezone('EST'))):
            y2019_holiday, y2019_cutday = BusyDayExchange.y2019()
        else:
            y2019_holiday, y2019_cutday = pandas.Series([], dtype=object), pandas.Series([], dtype=object)
        holiday = pandas.concat((y2020_holiday, y2019_holiday), ignore_index=True)
        cutday = pandas.concat((y2020_cutday, y2019_cutday), ignore_index=True)
        return holiday, cutday

    @staticmethod
    def get_holiday_range(start_date, end_date):
        holiday, _ = BusyDayExchange.get_ranges(start_date, end_date)
        return holiday

    @staticmethod
    def get_cutday_range(start_date, end_date):
        _, cutday = BusyDayExchange.get_ranges(start_date, end_date)
        return cutday

    @staticmethod
    def get_fullday_range(start_date, end_date):
        date_range = pandas.Series(pandas.date_range(start=start_date, end=end_date, freq='D'))
        holiday, cutday = BusyDayExchange.get_ranges(start_date, end_date)
        result = date_range[(~date_range.isin(holiday)) & (~date_range.isin(cutday))]
        return result

    @staticmethod
    def get_busyday_range(start_date, end_date):
        date_range = pandas.Series(pandas.date_range(start=start_date, end=end_date, freq='D'))
        result = date_range[date_range.apply(func=BusyDayExchange.is_holi)]
        return result

    @staticmethod
    def is_busy(day):
        start_day, end_day = BusyDayExchange.generate_start_end_dates(day)
        busy_range = BusyDayExchange.get_busyday_range(start_day, end_day)
        busy = (busy_range == day).sum() == 1
        return busy

    @staticmethod
    def is_full(day):
        start_day, end_day = BusyDayExchange.generate_start_end_dates(day)
        full_range = BusyDayExchange.get_fullday_range(start_day, end_day)
        full = (full_range == day).sum() == 1
        return full

    @staticmethod
    def is_holi(day):
        start_day, end_day = BusyDayExchange.generate_start_end_dates(day)
        holi_range = BusyDayExchange.get_holiday_range(start_day, end_day)
        holi = (holi_range == day).sum() == 1
        return holi

    @staticmethod
    def is_cut(day):
        start_day, end_day = BusyDayExchange.generate_start_end_dates(day)
        cut_range = BusyDayExchange.get_cutday_range(start_day, end_day)
        cut = (cut_range == day).sum() == 1
        return cut

    @staticmethod
    def to_holiday(day, *args):
        front = args[0]
        start_day, end_day = BusyDayExchange.generate_start_end_dates(day)
        holidays = BusyDayExchange.get_holiday_range(start_day, end_day)
        diff = holidays - day
        diff = diff.apply(func=lambda x: x.days)
        if front:
            result = holidays.iloc[diff.tolist().index(diff[diff > 0].min())]
        else:
            result = holidays.iloc[diff.tolist().index(diff[diff < 0].max())]
        result = (result - day).days
        return result

    @staticmethod
    def to_cutday(day, *args):
        front = args[0]
        start_day, end_day = BusyDayExchange.generate_start_end_dates(day)
        cutdays = BusyDayExchange.get_cutday_range(start_day, end_day)
        diff = cutdays - day
        diff = diff.apply(func=lambda x: x.days)
        if front:
            result = cutdays.iloc[diff.tolist().index(diff[diff > 0].min())]
        else:
            result = cutdays.iloc[diff.tolist().index(diff[diff < 0].max())]
        result = (result - day).days
        return result

    @staticmethod
    def to_fullday(day, *args):
        front = args[0]
        start_day, end_day = BusyDayExchange.generate_start_end_dates(day)
        fulldays = BusyDayExchange.get_fullday_range(start_day, end_day)
        diff = fulldays - day
        diff = diff.apply(func=lambda x: x.days)
        if front:
            result = fulldays.iloc[diff.tolist().index(diff[diff > 0].min())]
        else:
            result = fulldays.iloc[diff.tolist().index(diff[diff < 0].max())]
        result = (result - day).days
        return result

    @staticmethod
    def to_busyday(day, *args):
        front = args[0]
        start_day, end_day = BusyDayExchange.generate_start_end_dates(day)
        busydays = BusyDayExchange.get_busyday_range(start_day, end_day)
        diff = busydays - day
        diff = diff.apply(func=lambda x: x.days)
        if front:
            result = busydays.iloc[diff.tolist().index(diff[diff > 0].min())]
        else:
            result = busydays.iloc[diff.tolist().index(diff[diff < 0].max())]
        result = (result - day).days
        return result

    @staticmethod
    def generate_start_end_dates(day):
        return low_bound, top_bound


class BusyTimeExchange:

    @staticmethod
    def get_full_hours():
        start, end = datetime.time(hour=9, minute=30, second=0, microsecond=0,
                                   tzinfo=pytz.timezone('EST')), datetime.time(hour=16, minute=0, second=0,
                                                                               microsecond=0,
                                                                               tzinfo=pytz.timezone('EST'))
        return start, end

    @staticmethod
    def get_cut_hours():
        start, end = datetime.time(hour=9, minute=30, second=0, microsecond=0,
                                   tzinfo=pytz.timezone('EST')), datetime.time(hour=13, minute=0, second=0,
                                                                               microsecond=0,
                                                                               tzinfo=pytz.timezone('EST'))
        return start, end

    @staticmethod
    def get_holi_hours():
        start, end = pandas.NaT, pandas.NaT
        return start, end

    @staticmethod
    def get_open(day):
        result = None
        if BusyDayExchange.is_cut(day):
            result, _ = BusyTimeExchange.get_cut_hours()
        if BusyDayExchange.is_full(day):
            result, _ = BusyTimeExchange.get_full_hours()
        if BusyDayExchange.is_holi(day):
            result, _ = BusyTimeExchange.get_holi_hours()
        if not BusyDayExchange.is_holi(day) and result is not None:
            result = pandas.to_datetime(datetime.datetime(year=day.date().year, month=day.date().month, day=day.date().month,
                                                          hour=result.hour, minute=result.minute, tzinfo=result.tzinfo))
        return result

    @staticmethod
    def get_close(day):
        result = None
        if BusyDayExchange.is_cut(day):
            _, result = BusyTimeExchange.get_cut_hours()
        if BusyDayExchange.is_full(day):
            _, result = BusyTimeExchange.get_full_hours()
        if BusyDayExchange.is_holi(day):
            _, result = BusyTimeExchange.get_holi_hours()
        if not BusyDayExchange.is_holi(day) and result is not None:
            result = pandas.to_datetime(datetime.datetime(year=day.date().year, month=day.date().month, day=day.date().month,
                                                          hour=result.hour, minute=result.minute, tzinfo=result.tzinfo))
        return result

    @staticmethod
    def to_open(date):
        result = (BusyTimeExchange.get_open(date) - date)
        if pandas.isna(result):
            return pandas.NA
        else:
            return result.total_seconds()

    @staticmethod
    def to_close(date):
        result = (BusyTimeExchange.get_close(date) - date)
        if pandas.isna(result):
            return pandas.NA
        else:
            return result.total_seconds()
